/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

// motion model given left and right encoder counts and a starting
// pose.  computes a gaussian distribution for the new pose.

#include "ratbot-slam.h"

// FIXME: some of the constants being used here are really small (like
// c/2) and maybe it would make sense to change around the order of
// operations in some cases to reduce the error.  alternatively,
// change the units (e.g., to cm) and convert later on.

void motion_model(int16_t left, int16_t right, pose_t *mu, cov3_t *cov)
{
  // apply the motion model to obtain the new pose
  const fixed_t d = fp_mul(MOTION_c_over_2, int2fp(left)) + fp_mul(MOTION_c_over_2, int2fp(right));
  //  const fixed_t a = fp_mul(MOTION_c_over_2b, int2fp(right-left));
  const fixed_t a = fp_mul(MOTION_c_over_2b, int2fp(right)) - fp_mul(MOTION_c_over_2b, int2fp(left));
  mu->t = mu->t + a;
  const fixed_t ct = fp_cos(mu->t), st = fp_sin(mu->t);
  mu->x = mu->x + fp_mul(d, ct);
  mu->y = mu->y + fp_mul(d, st);

  // compute the new covariance: Px <- Jx*Px*Jx' + Ju*Pu*Ju'
  // be careful here, order of operations matters because we are
  // modifying cov directly

  // first, Jx*Px*Jx':
  const fixed_t Jx02 = fp_mul(-d, st), Jx12 = fp_mul(d, ct);
  const fixed_t tt_Jx02 = fp_mul(cov->tt, Jx02);
  const fixed_t tt_Jx12 = fp_mul(cov->tt, Jx12);
  cov->xx += fp_mul(Jx02,cov->xt) + fp_mul(Jx02, cov->xt + tt_Jx02);
  cov->xy += fp_mul(Jx12,cov->xt) + fp_mul(Jx02, cov->yt + tt_Jx12);
  cov->xt += tt_Jx02;
  cov->yy += fp_mul(Jx12,cov->yt) + fp_mul(Jx12, cov->yt + tt_Jx12);
  cov->yt += tt_Jx12;
  // cov->tt doesn't change

  // now add Ju*Pu*Ju': note that this assumes Pu is of the form:
  //  [sigma_l^2   0   ]
  //  [   0   sigma_r^2]
  // so if you want to use nonzero cross-covariance terms you'll have
  // to redo this!
  const fixed_t T1 = fp_mul(MOTION_c_over_2, ct);
  const fixed_t T2 = fp_mul(MOTION_c_over_2b, fp_mul(st, d));
  const fixed_t T3 = fp_mul(MOTION_c_over_2, st);
  const fixed_t T4 = fp_mul(MOTION_c_over_2b, fp_mul(ct, d));
  const fixed_t Ju00 = T1 - T2, Ju01 = T1 - T2, Ju10 = T3 + T4, Ju11 = T3 + T4;
  // Ju20 = c/2b, Ju21 = -c/2b
  fixed_t Pu_ll = fp_mul(VAR_ENCODER_PERC, int2fp(left)); Pu_ll = fp_mul(Pu_ll,Pu_ll);
  fixed_t Pu_rr = fp_mul(VAR_ENCODER_PERC, int2fp(right)); Pu_rr = fp_mul(Pu_rr,Pu_rr);

  cov->xx += fp_mul(Pu_ll, fp_mul(Ju00,Ju00)) + fp_mul(Pu_rr, fp_mul(Ju01,Ju01));
  cov->xy += fp_mul(Pu_ll, fp_mul(Ju00,Ju10)) + fp_mul(Pu_rr, fp_mul(Ju01,Ju11));
  cov->xt += fp_mul(Pu_rr, fp_mul(Ju01,MOTION_c_over_2b)) - fp_mul(Pu_ll, fp_mul(Ju00,MOTION_c_over_2b));
  cov->yy += fp_mul(Pu_ll, fp_mul(Ju10,Ju10)) + fp_mul(Pu_rr, fp_mul(Ju11,Ju11));
  cov->yt += fp_mul(Pu_rr, fp_mul(Ju11,MOTION_c_over_2b)) - fp_mul(Pu_ll, fp_mul(Ju10,MOTION_c_over_2b));
  cov->tt += fp_mul(fp_mul(Pu_ll + Pu_rr, MOTION_c_over_2b), MOTION_c_over_2b);
}
