/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

#include <sys/types.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/select.h>
#include <errno.h>
#include <inttypes.h>
#include <stdio.h>

static int ttyfd = -1;
static struct termios oldtios;
static const struct timeval to_default = {0, 0};
static struct timeval timeout = {0, 0};

int serial_shutdown(void);


/*!
  \brief Initialize serial communications with hardware.
*/
int serial_init(const char *mc_tty)
{
  struct termios tios;

  if(ttyfd > 0 && serial_shutdown() < 0)
    return -1;
  ttyfd = open(mc_tty, O_RDWR | O_NOCTTY);
  if(ttyfd < 0)
    return -1;

  /* save old settings */
  tcgetattr(ttyfd, &oldtios);

  memset(&tios, 0, sizeof(tios));
  tios.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
  tios.c_cc[VMIN] = 1; /* block until 1 char arrives */

  /* clean line, activate settings */
  if(tcflush(ttyfd, TCIFLUSH) < 0)
    goto init_error;
  if(tcsetattr(ttyfd, TCSANOW, &tios) < 0)
    goto init_error;

  return ttyfd;
init_error:
  timeout.tv_sec = 0;
  serial_shutdown();
  return -1;
}

/*!
  \brief Shutdown the hardware communications and the
  serial device they go through.
  \return -1 on failure, 0 on success
*/
int serial_shutdown(void)
{
  int rv = 0;
  /* restore old settings */
  if(tcsetattr(ttyfd, TCSANOW, &oldtios) < 0)
    rv = -1;
  if(close(ttyfd) < 0)
    rv = -1;
  ttyfd = -1;
  return rv;
}

static inline int wait_on_fd(int fd)
{
  if(timeout.tv_sec) {
    static fd_set r;
    static int n;
    FD_ZERO(&r);
    FD_SET(ttyfd, &r);
    n = select(ttyfd + 1, &r, 0, 0, &timeout);
    timeout = to_default; /* select changes it */
    return n;
  }
  return 1;
}

/*!
  \brief Read a hardware data packet from the serial device.
*/
int serial_read(uint8_t *out, int sz)
{
  if(ttyfd <= 0) {
    errno = EBADF;
    return -1;
  }
  if(sz != 0) {
    errno = 0;
    while(wait_on_fd(ttyfd) <= 0) {
      if(errno == EINTR)
	continue;
      if(errno == 0) errno = EILSEQ;
      return -1;
    }

    int n = 0;
    while(n != sz)
      n += read(ttyfd, out+n, sz-n);
  }

  return 0;
}

int serial_write(const uint8_t *data, int sz)
{
  if(ttyfd <= 0) {
    errno = EBADF;
    return -1;
  }
  if(sz != 0) {
    int n = write(ttyfd, data, sz);
    if(n != sz)
      return -1;
  }
  return 0;
}

int main(int argc, char **argv)
{
  if(argc < 2) {
    printf("Usage: %s logfile\n", argv[0]);
    return 1;
  }

  FILE *logfile = fopen(argv[1], "rb");
  if(!logfile) {
    printf("Unable to open %s\n", argv[1]);
    return 1;
  }

  if(serial_init("/dev/ttyUSB0") < 0) {
    perror("/dev/ttyUSB0");
    return 1;
  }

  uint32_t count = 0;
  int16_t frame[7];
  uint8_t ready;
  while(!feof(logfile)) {
    if(fread(frame, sizeof(int16_t), 7, logfile) != 7)
      break;
    printf("%d          \r", ++count); fflush(stdout);

    if(serial_write((uint8_t *)frame, 7*sizeof(int16_t)) < 0) {
      perror("serial_write()");
      break;
    }

    if(serial_read(&ready, 1) < 0) {
      perror("serial_read()");
      break;
    }
  }

  memset(frame, 0xff, 7*sizeof(int16_t));
  serial_write((uint8_t *)frame, 7*sizeof(int16_t));

  serial_shutdown();
  fclose(logfile);

  return 0;
}
